# Copyright (c) 2024 The KUTE contributors
import os 
import pytest
import numpy as np

from kute import GreenKuboIntegral, IntegralEnsemble
from itertools import chain, combinations
from kute.loaders import load_electric_current

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]




@pytest.fixture
def integral_ensemble(splits, replicas):

    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5")
    integrals = [GreenKuboIntegral(*load_electric_current(file, splits)) for _ in range(replicas)]

    return IntegralEnsemble(integrals)


class TestIntegralEnsemble(object):

    @pytest.mark.parametrize("splits,replicas", [(1, 2), (1, 5), (1, 10), (5, 2), (5, 5), (5, 10)])
    def test_load(self, integral_ensemble, replicas):

        assert integral_ensemble.N_REPLICAS == replicas

    @pytest.mark.parametrize("splits,replicas", [(1, 2), (1, 5), (1, 10), (5, 2), (5, 5), (5, 10)])
    def test_calculation(self, integral_ensemble):

        def capped_powerset(iterable):
            """Returns the set of all sets in the iterable that contain at least two elements
            """
            s = list(iterable)
            return list(chain.from_iterable(combinations(s, r) for r in range(2, len(s)+1)))

        keys = ["11", "22", "33", "12", "13", "23"]

        for key_set in capped_powerset(keys):

            t, avg, uavg = integral_ensemble.get_average_over_components(key_set)

            assert avg.shape == uavg.shape
            assert t.shape == avg.shape

    @pytest.mark.parametrize("splits,replicas", [(1, 2), (1, 5), (1, 10), (5, 2), (5, 5), (5, 10)])
    def test_specific_calculations(self, integral_ensemble):

        _, _, _ = integral_ensemble.get_isotropic_average()
        _, _, _ = integral_ensemble.get_anisotropic_average()


# Copyright (c) 2024 The KUTE contributors

import numpy as np
import h5py

from ._electriccurrentreporter import ElectricCurrentWriter
from kute import __version__
from openmm import unit
from typing import Union
import os
_PATHLIKE = Union[str, bytes, os.PathLike]


class SpatialElectricCurrentReporter(object):
    """
    Custom OpenMM reporter to calculate electric current in a given slice in the z direction

    Args:
        file (str): Name of the file to write the current to
        region (tuple): The values (zmin, zmax) that define the region to calculate the current in. They should be given in Angstroms
    """

    def __init__(self, file: _PATHLIKE, reportInterval: int, region: tuple):

        self._out = h5py.File(file, "w")
        self._writer = None
        self._totalSteps = None

        self._reportInterval = reportInterval

        self._zmin, self._zmax = region

    def __del__(self):
        self._out.close()

    def describeNextReport(self, simulation):
        """
        Function to be called by Openmm, gets information about the next report
        this object will generate.
        """
        steps = self._reportInterval - simulation.currentStep%self._reportInterval

        return (steps, True, True, False, False, True)     
    
    def report(self, simulation, state):
        """
        Function to be called by Openmm, generate a report.
        """
        if self._writer is None:
                self._writer = SpatialElectricCurrentWriter(self._out, simulation, self._zmin, self._zmax)

        positions = state.getPositions(True).value_in_unit(unit.angstrom)
        velocities = state.getVelocities(True).value_in_unit(unit.angstrom/unit.picosecond)
        time = state.getTime().value_in_unit(unit.picosecond)
        self._writer.writeCurrent(time, positions, velocities)


class SpatialElectricCurrentWriter(ElectricCurrentWriter):
     
    def __init__(self, file: h5py.File, simulation, zmin: float, zmax: float):
        """Class to write electric current in binary h5 format. Helper to SpatialElectricCurrentReporter

        Args:
            file (h5py.File): File where information will be written
            simulation : OpenMM object describing the simulation
            zmin (float): Minimum z value of the region to calculate the current in
            zmax (float): Maximum z value of the region to calculate the current in
        """

        self._zmin = zmin
        self._zmax = zmax
        super().__init__(file, simulation)
 
    def writeCurrent(self, time: np.ndarray, positions: np.ndarray, velocities: np.ndarray):
        """
        Function to calculate the electric current and write it to the file
        """
        z = positions[:, 2]
        mask = (z > self._zmin) & (z < self._zmax)
        current = np.sum(velocities[mask, :]*self._weight_vector[mask, np.newaxis], axis=0)

        LEN = self._file['timeseries/time'].shape[0]
        self._file['timeseries/time'].resize((LEN+1,))
        self._file['timeseries/time'][-1] = time
        self._file['timeseries/current'].resize((LEN+1,3))
        self._file['timeseries/current'][-1] = current
   


# Copyright (c) 2024 The KUTE contributors

from kute.analysis import ElectricCurrent, COMVelocity, PressureTensor
import MDAnalysis
from MDAnalysis.analysis.base import AnalysisBase
import argparse
from ._ascii_logo import print_logo


def get_analysis_class(property:str) -> callable:
    """
    Given a string describing a current, return the corresponding analysis class.

    Args:
        property (str): String describing the current. Currently supported currents are "electric_current", "com_velocity" and "pressure_tensor".

    Raises:
        ValueError: If a current is requested that is not supported

    Returns:
        callable: A class from kute.analysis that calculates the appropriate microscopic current
    """
    equivalence_dictionary ={"electric_current":ElectricCurrent, 
                             "com_velocity": COMVelocity, 
                             "pressure_tensor": PressureTensor}

    if property in equivalence_dictionary:
        return equivalence_dictionary[property]
    
    else:
        raise ValueError("The requested property is not supported. Supported options are " + " ".join(equivalence_dictionary.keys()))


def perform_analysis(universe: MDAnalysis.Universe, 
                     analysis_class: AnalysisBase, 
                     filename: str) -> None:

    
    analysis = analysis_class(universe, filename)
    analysis.run(verbose=True)


def main():

    description = "Calculate microscopic currents from MD trajectories. When using this utility, make sure that the selected trajectories contain the appropriate information at the same recording frequencies. The requirements are as follows: Electric current: Topology must include bonds, masses and charges, trajectory must include velocities. Center of mass velocity: Topology must include bonds and masses, trajectory must include velocities. Pressure tensor: Topology must include masses, trajectory must include positions, velocities and forces."

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("--current", type=str, required=True, metavar="electric_current", help="Type of current to be calculated. Supported options are 'electric_current', 'com_velocity' and 'pressure_tensor'", dest="current")

    parser.add_argument("-f", "--traj", type=str, help="Path to the trajectory file (.trr, .xtc, .dcd, etc.)", metavar="run.trr", required=True, dest="traj")

    parser.add_argument("-t", "--top", type=str, help="Path to the topology file (.gro, .pdb, .tpr, etc.)", metavar="run.tpr", required=True, dest="top")

    parser.add_argument("-o", "--output", type=str, help="Name of the output file", metavar="current.h5", required=False, default="current.h5", dest="output")

    

    args = parser.parse_args()
    print_logo()

    u = MDAnalysis.Universe(args.top, args.traj)
    analysis_class = get_analysis_class(args.current)

    perform_analysis(u, analysis_class, args.output)
    


# Copyright (c) 2024 The KUTE contributors

import argparse
import numpy as np
import MDAnalysis as mda

from typing import Tuple
from ._ascii_logo import print_logo
from kute import GreenKuboIntegral, IntegralEnsemble
from kute.loaders import load_electric_current, load_com_velocity, load_pressure_tensor



def get_loader_function(string:str, resname:str=None) -> callable:
    """Given an input string describing a transport coefficient, return the corresponding loader function.

    Args:
        string (str): String describing the transport coefficient
        resname (str, optional): Name of the residue for which to calculate the diffusion coefficient. Defaults to None.

    Raises:
        ValueError: If a transport coefficient is requested that is not supported

    Returns:
        callable: A function from kute.loaders that loads the appropriate microscopic current
    """

    if string == "diffusion_coefficient" and resname is None:
        raise ValueError("The residue name is required to calculate the diffusion coefficient")

    equivalence_dictionary = {"electric_conductivity": load_electric_current, 
                              "diffusion_coefficient": lambda filename, splits: load_com_velocity(file=filename, resname=resname, splits=splits), 
                              "viscosity": load_pressure_tensor}

    if string in equivalence_dictionary:
        return equivalence_dictionary[string]
    
    else:
        raise ValueError("The requested transport coefficient is not supported. Supported options are " + " ".join(equivalence_dictionary.keys()))
    

def calculation_for_one(filename:str, loader:callable, splits:int, weight_file:str=None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Runs the calculation for a single current file

    Args:
        filename (str): Name of the .h5 file containing the microscopic current
        loader (callable): A function from kute.loaders that loads the appropriate microscopic current
        splits (int): Number of splits to use for the Green-Kubo integral
        weight_file (str, optional): Name of the file containing extra multiplicative factors. Defaults to None.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: lag time, isotropic average running transport coefficient, uncertainty
    """

    if weight_file is not None:
        weight = np.loadtxt(weight_file)
    else:
        weight = 1

    t, Jx, Jy, Jz = loader(filename, splits)
    integral = GreenKuboIntegral(t, Jx, Jy, Jz)
    tavg, avg, u_avg = integral.get_isotropic_running_average()

    avg *= weight
    u_avg *= weight

    return tavg, avg, u_avg


def calculation_for_ensemble(filenames:iter, loader:callable, splits:int, weight_file:str=None)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Runs the calculation for any number of .h5 files, treating them as an ensemble

    Args:
        filename (iter): An iterable containing the names of the .h5 files containing the microscopic current
        loader (callable): A function from kute.loaders that loads the appropriate microscopic current
        splits (int): Number of splits to use for the Green-Kubo integral
        weight_file (str, optional): Name of the file containing extra multiplicative factors. Defaults to None.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: lag time, isotropic average transport coefficient, uncertainty
    """

    if weight_file is not None:
        weights = np.loadtxt(weight_file)
    else:
        weights = weight_file

    integrals = []
    for name in filenames:
        t, Jx, Jy, Jz = loader(name, splits)
        integrals.append(GreenKuboIntegral(t, Jx, Jy, Jz))
    
    ensemble = IntegralEnsemble(integrals, factors=weights)
    return ensemble.get_isotropic_average()


def save_results(out_file:str, tavg:np.ndarray, avg:np.ndarray, uavg:np.ndarray):
    """Save the results of the calculation to a file

    Args:
        out_file (str): Name of the file to save the results to
        tavg (np.ndarray): Lag time
        avg (np.ndarray): Isotropic average transport coefficient
        uavg (np.ndarray): Uncertainty
    """

    to_save = np.vstack([tavg, avg, uavg]).T
    np.savetxt(out_file, to_save, header="Time       Average         Uncertainty")


def main():

    description = "Calculates the isotropic transport coefficient as a function of averaging cutoff"

    parser = argparse.ArgumentParser(description=description)

    ## Input arguments

    parser.add_argument("-f", required=True, type=str, dest="h5_files", metavar="current.h5", nargs="+", help = "List of h5 binary files containing the microscopic current for each replica.")

    parser.add_argument("--splits", required=False, type=int, dest="splits", metavar="1", default=1, help = "Number of fragments of equal size into which to split the pressure tensor of each replica")

    parser.add_argument("--weights", required=False, type=str, dest="weights", metavar="weights.txt", default=None, help = "File containing the weighting factor for each replica. Can be used to change units or to include replica-dependent values such as the volume")

    parser.add_argument("--property", required=True, type=str, dest="property", metavar="electric_conductivity", help = "The transport coefficient to calculate. Supported options are electric_conductivity, diffusion_coefficient and viscosity")

    parser.add_argument("--resname", required=False, type=str, dest="resname", metavar="resname", default=None, help = "Name of the residue for which to calculate the diffusion coefficient. Not required for the other transport coefficients")

    ## Output arguments

    parser.add_argument("-o", "--output", required=False, type=str, dest="output", metavar="output.dat", default="transport_coefficient.dat", help="Name of the file to save the results to")

    ## Load the arguments, print the KUTE logo

    args = parser.parse_args()
    print_logo()

    ## Get the loader function. In the case of diffusion coefficient, modify it to include the resname argument

    loader = get_loader_function(args.property, args.resname)

    ## Run the calculation

    if len(args.h5_files) == 1:
        tavg, avg, uavg = calculation_for_one(args.h5_files[0], loader, args.splits, args.weights)

    else:
        tavg, avg, uavg = calculation_for_ensemble(args.h5_files, loader, args.splits, args.weights)

    ## Save the results

    save_results(args.output, tavg, avg, uavg)

if __name__ == "__main__":
    main()

            



# Copyright (c) 2024 The KUTE contributors

import numpy as np
import argparse
import shutil
import h5py

from warnings import warn
from ._ascii_logo import print_logo
from kute import __version__
from typing import Union
import os
_PATHLIKE = Union[str, bytes, os.PathLike]

def join_h5_files(output_file_path: _PATHLIKE, ignore_checks, *file_paths: _PATHLIKE) -> None:
    """Join two or more kute h5 files into a single file.

    Args:
        output_file_path (str): Path to the output file
        check
        file_paths (str): Paths to the files to be joined
    """

    # First check if more than one file is being joined
    if len(file_paths) < 2:
        raise ValueError("At least two files must be provided to join.")

    # Duplicate the first file and rename as the output file
    try:
        shutil.copy(file_paths[0], output_file_path)
    except shutil.SameFileError:
        pass

    # Open the output file in append mode
    with h5py.File(output_file_path, 'a') as output_file:

        # Get the attributes of the first file
        ff_version = output_file['information'].attrs['kute_version']
        time_units = output_file['information/units'].attrs['time']
        current_units = next(output_file['information/units'].attrs[key] for key in output_file['information/units'].attrs if key != 'time')
        
        # Check compatibility between files
        for file in file_paths[1:]:

            with h5py.File(file, 'r') as input_file:
                
                info_sec = input_file['information']
                
                if not ignore_checks:

                    # Check if the version of the files is the same
                    if info_sec.attrs['kute_version'] != ff_version:
                        warn("The kute version of the files is not the same. This could cause incompatibility issues.")
                    
                    # Check if the files have the same time units
                    if info_sec['units'].attrs['time'] != time_units:
                        # Close the output file and remove it
                        output_file.close()
                        os.remove(output_file_path)
                        raise ValueError("The time units of the files are not the same.")
                    
                    # Check if the files have the same current units
                    if next(info_sec['units'].attrs[key] for key in info_sec['units'].attrs if key != 'time') != current_units:
                        # Close the output file and remove it
                        output_file.close()
                        os.remove(output_file_path)
                        raise ValueError("The current units of the files are not the same.")

        current_key = next(key for key in output_file['timeseries'].keys() if  key != 'time')

        # Join the files
        for file in file_paths[1:]:
            
            with h5py.File(file, 'r') as input_file:
               
               # Combine the timseries datasets
               output_time = output_file['timeseries/time']
               input_time = input_file['timeseries/time']

               input_time = np.array(input_time[:].copy())

               if output_time[-1] == input_time[0]:
                    # If the first time is equal to the last time of the previous file, dont repeat it
                    output_time.resize((output_time.shape[0] + input_time.shape[0] - 1,))
                    output_time[-(input_time.shape[0]-1):] = input_time[1:]

               else:
                    output_time.resize((output_time.shape[0] + input_time.shape[0],))
                    output_time[-input_time.shape[0]:] = input_time[:]
               
               # Combine current datasets
               output_current = output_file[f'timeseries/{current_key}']
               try:
                   input_current = input_file[f'timeseries/{current_key}']
               except KeyError:
                    # Close the output file and remove it
                    output_file.close()
                    os.remove(output_file_path)
                    raise ValueError("The files do not have the same current datasets.")
               
               if len(output_current.shape) == 1:
                   output_current.resize((output_current.shape[0] + input_current.shape[0],))
               else:
                   new_size = list(output_current.shape)
                   new_size[0] += input_current.shape[0]
                   output_current.resize(tuple(new_size))
                    
               output_current[-input_current.shape[0]:] = input_current[:]
                               

def main():

    description = "Join two or more kute h5 files into a single file."

    parser = argparse.ArgumentParser(description=description)

    ## Input arguments

    parser.add_argument("-o", "--output", required=True, type=str, dest="output_file", metavar="output.h5", help = "Name of the output h5 file with the joined data")

    parser.add_argument("-f", required=True, type=str, dest="input_files", metavar="current.h5", nargs="+", help = "List of h5 binary files to join")

    parser.add_argument("--ignore_checks", action="store_true", default=False, dest="ignore_checks", help="Ignore compatibility checks between files")

    ## Load the arguments, print the KUTE logo

    args = parser.parse_args()
    print_logo()

    ## Join the files

    join_h5_files(args.output_file, args.ignore_checks, *args.input_files)

    print(f"Joined files saved to {args.output_file}.")

if __name__ == "__main__":
    main()

            
                






# Copyright (c) 2024 The KUTE contributors

import os
import pytest
import numpy as np
import kute.__main__
import sys
import MDAnalysis as mda
import h5py

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def microscopic_current_file(property):
    equivalences = {"electric_conductivity":  os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5"),
                    "diffusion_coefficient": os.path.join(ACTUAL_PATH, "files_for_tests/inputs/com_velocity_test.h5"),
                    "viscosity": os.path.join(ACTUAL_PATH, "files_for_tests/inputs/pressure_tensor_test.h5")}
    
    return equivalences[property]


@pytest.fixture
def property(property_name):
    return property_name


class TestCalculateTransportCoefficientRoutine(object):

    @pytest.mark.parametrize("property_name,resname", [("electric_conductivity", None), ("diffusion_coefficient", "ea"), ("diffusion_coefficient", "no3"),("viscosity", None)])
    def test_loader_functions(self, property, microscopic_current_file, resname):
        
        from kute.routines._calculate_transport_coefficient import get_loader_function

        loader = get_loader_function(property, resname)
        t, Jx, Jy, Jz = loader(microscopic_current_file, 1)


    @pytest.mark.parametrize("property_name", ["electric_conductivity", "diffusion_coefficient", "viscosity"])
    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    @pytest.mark.parametrize("resname", ["ea", "no3"])
    def test_calculation_for_one(self, property, microscopic_current_file, resname, splits):
        
        from kute.routines._calculate_transport_coefficient import calculation_for_one
        from kute.routines._calculate_transport_coefficient import get_loader_function    

        loader = get_loader_function(property, resname)

        ## Create random weight for the test

        weight = np.array(np.random.random((1,1)))
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_one(microscopic_current_file, loader, splits, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")


    @pytest.mark.parametrize("property_name", ["electric_conductivity", "diffusion_coefficient", "viscosity"])
    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    @pytest.mark.parametrize("resname", ["ea", "no3"])
    @pytest.mark.parametrize("repetitions", [5, 10])
    def test_calculation_for_ensemble(self, property, microscopic_current_file, resname, splits, repetitions):

        from kute.routines._calculate_transport_coefficient import calculation_for_ensemble
        from kute.routines._calculate_transport_coefficient import get_loader_function

        loader = get_loader_function(property, resname)

        weight = np.random.random(repetitions) 
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_ensemble([microscopic_current_file for _ in range(repetitions)], loader, splits, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")

    @pytest.mark.parametrize("number", [100, 500, 1000])
    def test_save_results(self, number):
    
        from kute.routines._calculate_transport_coefficient import save_results

        t, avg, uavg = np.random.random((3, number))
        save_results("temporal.dat", t, avg, uavg)

        assert os.path.isfile("temporal.dat")

        os.remove("temporal.dat")

    @pytest.mark.parametrize("property_name", ["electric_conductivity", "diffusion_coefficient", "viscosity"])
    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    @pytest.mark.parametrize("resname", ["ea", "no3"])
    @pytest.mark.parametrize("repetitions", [5, 10])
    def test_commandline_routine(self, microscopic_current_file, property, splits, repetitions, resname):
        
        file_list = " ".join([microscopic_current_file for _ in range(repetitions)])
        command = f"kute calculate_transport_coefficient -f {file_list} --splits {splits} --output temporary_out.dat --resname {resname} --property {property}"

        sys.argv = command.split()
        kute.__main__.main()

        assert os.path.isfile("temporary_out.dat")
        os.remove("temporary_out.dat")


@pytest.fixture
def trajectory_file():
    return os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")

@pytest.fixture
def topology_file():
    return os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")

class TestCalculateMicroscopicCurrentRoutine(object):

    @pytest.mark.parametrize("current_type", ["electric_current", "com_velocity", "pressure_tensor"])
    def test_get_analysis_class(self, current_type):

        from kute.routines._calculate_microscopic_current import get_analysis_class

        analysis_class = get_analysis_class(current_type)

    @pytest.mark.parametrize("current_type", ["electric_current", "com_velocity", "pressure_tensor"])
    def test_perform_analysis(self, trajectory_file, topology_file, current_type):

        from kute.routines._calculate_microscopic_current import get_analysis_class, perform_analysis

        u = mda.Universe(topology_file, trajectory_file)
        analysis_class = get_analysis_class(current_type)

        perform_analysis(u, analysis_class, "temporary_out.h5")

        assert os.path.isfile("temporary_out.h5")
        os.remove("temporary_out.h5")


    @pytest.mark.parametrize("current_type", ["electric_current", "com_velocity", "pressure_tensor"])
    def test_commandline_routine(self, trajectory_file, topology_file, current_type):

        command = f"kute calculate_microscopic_current --current {current_type} --traj {trajectory_file} --top {topology_file} --output temporary_out.h5"

        sys.argv = command.split()
        kute.__main__.main()

        assert os.path.isfile("temporary_out.h5")
        os.remove("temporary_out.h5")


@pytest.fixture
def current_file(prop):
    equivs = {"electric_current": os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5"),
              "com_velocity": os.path.join(ACTUAL_PATH, "files_for_tests/inputs/com_velocity_test.h5"),
              "pressure_tensor": os.path.join(ACTUAL_PATH, "files_for_tests/inputs/pressure_tensor_test.h5")}
    
    return equivs[prop]


class TestJoinCurrentsRoutine(object):

    @pytest.mark.parametrize("prop", ["electric_current", "com_velocity", "pressure_tensor"])
    @pytest.mark.parametrize("repetitions", [2, 5, 10])
    @pytest.mark.parametrize("CHECK", [True, False])
    def test_join_h5_files(self, repetitions, current_file, CHECK):

        from kute.routines._join_currents import join_h5_files

        file_list = [current_file for _ in range(repetitions)]

        join_h5_files("temporary_out.h5", CHECK, *file_list)

        assert os.path.isfile("temporary_out.h5")

        fold = h5py.File(current_file)
        L_old = len(np.array(fold['timeseries/time']))

        fnew = h5py.File("temporary_out.h5")
        L_new = len(np.array(fnew['timeseries/time']))

        assert L_new == L_old * repetitions

        os.remove("temporary_out.h5")

    @pytest.mark.parametrize("prop", ["electric_current", "com_velocity", "pressure_tensor"])
    @pytest.mark.parametrize("repetitions", [2, 5, 10])
    @pytest.mark.parametrize("CHECK", [True, False])
    def test_commandline_routine(self, repetitions, current_file, CHECK):

        file_list =" ".join([current_file for _ in range(repetitions)])
        command = f"kute join_currents -f {file_list} -o temporary_out.h5"
        if CHECK:
            command += " --ignore_checks"

        sys.argv = command.split()
        kute.__main__.main()

        assert os.path.isfile("temporary_out.h5")
        os.remove("temporary_out.h5")







# Copyright (c) 2024 The KUTE contributors

import os 
import pytest
import getpass 
import h5py
import numpy as np

from kute import __version__
from kute.loaders import load_electric_current, load_com_velocity, load_pressure_tensor

try: 
    import openmm
    from openmm import unit
    import openmm.app as app

    from kute.reporters import ElectricCurrentReporter, COMVelocityReporter, PressureTensorReporter, SpatialElectricCurrentReporter

    SKIP_TESTS = False

except ImportError:
    SKIP_TESTS = True


ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def simulation():

    FIELD = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/field_openmm.xml")
    CONFIG = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/config_openmm.pdb")
    
    TEMP = 300 * unit.kelvin
    TIMESTEP = 1.0 * unit.femtosecond

    forcefield = app.ForceField(FIELD)
    pdb = app.PDBFile(CONFIG)

    system = forcefield.createSystem(pdb.topology,
                                     nonbondedMethod=app.PME,
                                     nonbondedCutoff=12.0*unit.angstrom,
                                     constraints=app.HBonds,
                                     ewaldErrorTolerance=1.0e-5)

    integrator = openmm.NoseHooverIntegrator(TEMP, 10 / unit.picoseconds, TIMESTEP)
    sim = app.Simulation(pdb.topology, system, integrator)
    sim.context.setPositions(pdb.positions)
    sim.context.setVelocitiesToTemperature(TEMP)

    return sim


class TestElectricCurrentReporter(object):

    @pytest.mark.skipif(SKIP_TESTS, reason="OpenMM is not installed")
    def test_reporter(self, simulation):

        simulation.reporters.clear()
        
        current_reporter = ElectricCurrentReporter(os.path.join(ACTUAL_PATH, "report_current_temp.h5"), reportInterval=1)
        simulation.reporters.append(current_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_current_temp.h5")
        assert os.path.isfile(FILEPATH)

        ## Check if the file can be loaded with KUTE

        t, Jx, Jy, Jz = load_electric_current(FILEPATH)
        
        assert Jx.shape == (1, len(t))
        assert Jy.shape == (1, len(t))
        assert Jz.shape == (1, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH, "r") as f:

            # Check general information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["electric_current"] == "e * A / ps"

            # Check for correct shapes

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/current"]).shape == (TOTALSTEPS, 3)

        os.remove(FILEPATH)

class TestSpatialElectricCurrentReporter(object):

    @pytest.mark.skipif(SKIP_TESTS, reason="OpenMM is not installed")
    @pytest.mark.parametrize("region", [(1, 2), (3, 4), (5, 6)])
    def test_reporter(self, simulation, region):

        simulation.reporters.clear()
        
        current_reporter = SpatialElectricCurrentReporter(os.path.join(ACTUAL_PATH, "report_current_temp.h5"), reportInterval=1, region=region)
        simulation.reporters.append(current_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_current_temp.h5")
        assert os.path.isfile(FILEPATH)

        ## Check if the file can be loaded with KUTE

        t, Jx, Jy, Jz = load_electric_current(FILEPATH)
        
        assert Jx.shape == (1, len(t))
        assert Jy.shape == (1, len(t))
        assert Jz.shape == (1, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH, "r") as f:

            # Check general information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["electric_current"] == "e * A / ps"

            # Check for correct shapes

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/current"]).shape == (TOTALSTEPS, 3)

        os.remove(FILEPATH)

class TestCOMVelocityReporter(object):

    @pytest.mark.skipif(SKIP_TESTS, reason="OpenMM is not installed")
    def test_reporter(self, simulation):

        simulation.reporters.clear()
        
        vel_reporter = COMVelocityReporter(os.path.join(ACTUAL_PATH, "report_com_vel_temp.h5"), reportInterval=1)
        simulation.reporters.append(vel_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_com_vel_temp.h5")
        assert os.path.isfile(FILEPATH)

        # Check if it can be loaded with KUTE

        names = np.unique([r.name for r in simulation.topology.residues()])
        for name in names:
            t, Vx, Vy, Vz = load_com_velocity(FILEPATH, name)
            number = np.sum([ 1 if r.name==name else 0 for r in simulation.topology.residues() ])
            assert Vx.shape == (number, len(t))
            assert Vy.shape == (number, len(t))
            assert Vz.shape == (number, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH) as f:

            # General information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["com_velocities"] == "A / ps"

            # Check residue counts

            for name in names:

                number = np.sum([ 1 if r.name==name else 0 for r in simulation.topology.residues() ])
                assert len(np.array(f[f"residues/{name}"])) == number

            # Check the shape of the results
            
            target_shape = (TOTALSTEPS, simulation.topology.getNumResidues(), 3)

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/com_velocities"]).shape == target_shape

        
        os.remove(FILEPATH)

class TestPressureTensorReporter(object):

    @pytest.mark.skipif(SKIP_TESTS, reason="OpenMM is not installed")
    def test_reporter(self, simulation):

        simulation.reporters.clear()
        
        current_reporter = PressureTensorReporter(os.path.join(ACTUAL_PATH, "report_pressure_tensor_temp.h5"), reportInterval=1)
        simulation.reporters.append(current_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_pressure_tensor_temp.h5")
        assert os.path.isfile(FILEPATH)

        ## Check if the file can be loaded with KUTE

        t, Pxy, Pxz, Pyz = load_pressure_tensor(FILEPATH)
        
        assert Pxy.shape == (1, len(t))
        assert Pxz.shape == (1, len(t))
        assert Pyz.shape == (1, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH, "r") as f:

            # Check general information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["pressure_tensor"] == "Pa"

            # Check for correct shapes

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/pressure_tensor"]).shape == (TOTALSTEPS, 3)

        os.remove(FILEPATH)


# Copyright (c) 2024 The KUTE contributors

import numpy as np
import h5py
import getpass

from kute import __version__
from MDAnalysis.analysis.base import AnalysisBase
from MDAnalysis import Universe
from MDAnalysis.analysis.base import Results


class COMVelocity(AnalysisBase):
    """
    Class to calculate center of mass velocities from MD trajectories.

    Args:
        universe (MDAnalysis.Universe): Universe containig the simulation
        filename (str, optional): Name of the h5 file to which the velocities will be saved. 
                                  Defaults to "com_velocity.h5".
    """
    def __init__(self, universe: Universe, filename:str="com_velocity.h5", **kwargs):

        super().__init__(universe.trajectory, **kwargs)
        self.u = universe
        self.filename = filename
        self.results = Results()

    
    def _prepare(self):

        atom_residue_masses = np.array([ a.residue.mass for a in self.u.atoms ])
        atom_masses = self.u.atoms.masses
        weights = atom_masses / atom_residue_masses
        self._matrix = np.zeros((self.u.residues.n_residues, self.u.atoms.n_atoms))

        for i, res in enumerate(self.u.residues):
            for atom in res.atoms:
                j = atom.index
                self._matrix[i, j] = weights[j]

        self.results.com_vel = np.zeros((self.n_frames, self.u.residues.n_residues, 3))

    def _single_frame(self):

        self.results.com_vel[self._frame_index, :, :] = self._matrix @ self.u.atoms.velocities
    
    def _conclude(self):

        self.write_h5_file()

    def write_h5_file(self):

        with h5py.File(self.filename, "w") as file:

        ## Metadata group

            file.create_group('information')
            file['information'].attrs['kute_version'] = __version__
            file['information'].attrs['author'] = getpass.getuser()
            file['information'].create_group('units')
            file['information/units'].attrs['time'] = "ps"
            file['information/units'].attrs['com_velocities'] = "A / ps"

            ## Residue identificators

            file.create_group('residues')
            names = self.u.residues.resnames
            for name in np.unique(names):
                where = np.where(names==name)[0]
                file['residues'].create_dataset(name, data=where, dtype=int)

            ## Data group

            file.create_group('timeseries')
            file['timeseries'].create_dataset('time', data=self.times, maxshape=(None,))
            file['timeseries'].create_dataset('com_velocities', data=self.results.com_vel, maxshape=(None, self.u.residues.n_residues, 3))


# Copyright (c) 2024 The KUTE contributors

# Constants (in S.I. units)
NA = 6.02214076e23 # Avogadro's number
KB = 1.380649e-23 # Boltzmann constant in J/K
EPSILON_0 = 8.8541878128e-12 # Vacuum permittivity in F/m
ELEMENTARY_CHARGE = 1.602176634e-19 # Elementary charge in Coulombs

# Transformations
KG_TO_AMU = 1 / 1.66053906660e-27 # Kilograms to atomic mass units
AMU_TO_KG = 1 / KG_TO_AMU
J_TO_KCAL = 1 / 4184 # Joules to kilocalories
KCAL_TO_J = 1 / J_TO_KCAL
J_TO_ERGS = 1e7 # Joules to ergs
ERGS_TO_J = 1 / J_TO_ERGS
J_TO_EV = 1 / 1.602176634e-19 # Joules to electron volts
EV_TO_J = 1 / J_TO_EV
METERS_TO_NANOMETERS = 1e9 # Meters to nanometers
NANOMETERS_TO_METERS = 1  / METERS_TO_NANOMETERS
METERS_TO_ANGSTROMS = 1e10 # Meters to angstroms
ANGSTROMS_TO_METERS = 1 / METERS_TO_ANGSTROMS
SEC_TO_PS = 1e12 # Seconds to picoseconds
PS_TO_SEC = 1 / SEC_TO_PS
SEC_TO_FS = 1e15 # Seconds to femtoseconds
FS_TO_SEC = 1 / SEC_TO_FS
PAS_TO_POISE = 1e1 # Pascal-seconds to 
POISE_TO_PAS = 1 / PAS_TO_POISE
NEWTONS_TO_DYNES = 1e5 # Newtons to dynes
DYNES_TO_NEWTONS = 1 / NEWTONS_TO_DYNES
PA_TO_PSI = 0.000145038 # Pascals to pounds per square inch
PSI_TO_PA = 1 / PA_TO_PSI

# Atomic masses (in atomic mass units)
ATOMIC_MASSES = {
    "H": 1.00784,
    "He": 4.0026,
    "Li": 6.9410,
    "Be": 9.0121831,
    "B": 10.811,
    "C": 12.011,
    "N": 14.007,
    "O": 15.999,
    "F": 18.998,
    "Ne": 20.180,
    "Na": 22.990,
    "Mg": 24.305,
    "Al": 26.982,
    "Si": 28.086,
    "P": 30.974,
    "S": 32.065,
    "Cl": 35.453,
    "Ar": 39.948,
    "K": 39.098,
    "Ca": 40.078,
    "Sc": 44.956,
    "Se": 78.960,
    "Ti": 47.867,
    "Br": 79.904,
    "Ag": 107.87,
    "Au": 196.967,
    "Hg": 200.59,
    "Rn": 222.0,
}


# Copyright (c) 2024 The KUTE contributors

import numpy as np
import h5py
from typing import Union, Optional, Tuple
import os
_PATHLIKE = Union[str, bytes, os.PathLike]

def load_electric_current(file: _PATHLIKE, 
                          splits:Optional[int]=1) -> Tuple[np.ndarray, np.ndarray,
                                                              np.ndarray, np.ndarray]:
    """
    Function to load electric currents from h5 files generated with KUTE.

    Args:
        file (str): Path to the h5 file
        splits (int, optional): Number of equal intervals into which the current will be split. Defaults to one.

    Returns:
        t, Jx, Jy, Jz: Time and xyz components of the electric current, given as
                       a proper input to the kute.GreenKuboIntegral class. See
                       that class documentation for more information.
    """

    with h5py.File(file) as f:
        time = np.array(f["timeseries/time"])
        Jx, Jy, Jz = np.array(f["timeseries/current"]).T

    time = np.split(time, splits)[0]
    Jx = np.array(np.split(Jx, splits))
    Jy = np.array(np.split(Jy, splits))
    Jz = np.array(np.split(Jz, splits))

    return time, Jx, Jy, Jz
 

def load_pressure_tensor(file: _PATHLIKE, 
                         splits:Optional[int]=1) -> Tuple[np.ndarray, np.ndarray,
                                                              np.ndarray, np.ndarray]:
    """
    Function to load off-diagonal components of the pressure tensor from h5 files generated with KUTE.

    Args:
        file (str): Path to the h5 file
        splits (int, optional): Number of equal intervals into which the tensor will be split. Defaults to one.

    Returns:
        t, Pxy, Pxz, Pyz: Time and off-diagonal components of the pressure tensor, given as
                          a proper input to the kute.GreenKuboIntegral class. See
                          that class documentation for more information.
    """

    with h5py.File(file) as f:
        time = np.array(f["timeseries/time"])
        Pxy, Pxz, Pyz = np.array(f["timeseries/pressure_tensor"]).T

    time = np.split(time, splits)[0]
    Pxy = np.array(np.split(Pxy, splits))
    Pxz = np.array(np.split(Pxz, splits))
    Pyz = np.array(np.split(Pyz, splits))

    return time, Pxy, Pxz, Pyz

def load_com_velocity(file: _PATHLIKE, resname: str, 
                      splits: Optional[int]=1) -> Tuple[np.ndarray, np.ndarray,
                                                           np.ndarray, np.ndarray]:
    """
    Function to load center of mass velocities for residues from h5 files generated with KUTE.

    Args:
        file (str): Path to the h5 file
        resname (str): Name of the residue for which the center of mass velocity will be loaded.
        splits (int, optional): Number of equal intervals into which the velocity will be split. Defaults to one.

    Returns:
        t, Jx, Jy, Jz: Time and xyz components of the center of mass velocity, given as a proper input to the
                       kute.GreenKuboIntegral class. See that class documentation for more information.

    Raises:
        ValueError: If the requested residue is not present in the file.
    """
    with h5py.File(file) as f:

        if resname not in f["residues"].keys():
            raise ValueError("Residue not present. Available residues are: " + ", ".join(list(f["residues"].keys())))
        
        mask = np.array(f[f"residues/{resname}"])
        time = np.array(f["timeseries/time"])
        Vx, Vy, Vz = np.array(f["timeseries/com_velocities"])[:, mask, :].T

    time = np.split(time, splits)[0]
    Vx = np.vstack(np.split(Vx, splits, axis=1))
    Vy = np.vstack(np.split(Vy, splits, axis=1))
    Vz = np.vstack(np.split(Vz, splits, axis=1))


    return time, Vx, Vy, Vz


# Copyright (c) 2024 The KUTE contributors

__version__ = "1.2.5"

from ._kute import GreenKuboIntegral, IntegralEnsemble
from . import analysis
from . import loaders
from . import constants

__all__ = ["GreenKuboIntegral", "IntegralEnsemble", "analysis", "loaders", "constants"]



# Copyright (c) 2024 The KUTE contributors

from ._com_velocity import COMVelocity
from ._electric_current import ElectricCurrent
from ._pressure_tensor import PressureTensor

__all__ = ["ElectricCurrent", "COMVelocity", "PressureTensor"]


# Copyright (c) 2024 The KUTE contributors

import os 
import pytest
import numpy as np

from kute import GreenKuboIntegral
from kute.loaders import load_electric_current
from itertools import chain, combinations

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def current(splits):
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5")
    return load_electric_current(file, splits)

@pytest.fixture
def green_kubo_integral(splits):
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5")
    return GreenKuboIntegral(*load_electric_current(file, splits))

@pytest.fixture
def results_correlation():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_correlation_test.dat")
    c11, c22, c33, c12, c13, c23 = np.loadtxt(file).T
    return c11, c22, c33, c12, c13, c23 

@pytest.fixture
def uncertainty_correlation():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/uncertainty_correlation_test.dat")
    uc11, uc22, uc33, uc12, uc13, uc23 = np.loadtxt(file).T
    return uc11, uc22, uc33, uc12, uc13, uc23 

@pytest.fixture
def results_integral():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_integral_test.dat")
    c11, c22, c33, c12, c13, c23 = np.loadtxt(file).T
    return c11, c22, c33, c12, c13, c23 

@pytest.fixture
def uncertainty_integral():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/uncertainty_integral_test.dat")
    uc11, uc22, uc33, uc12, uc13, uc23 = np.loadtxt(file).T
    return uc11, uc22, uc33, uc12, uc13, uc23 

@pytest.fixture
def results_running_average():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_running_average_test.dat")
    c11, c22, c33, c12, c13, c23 = np.loadtxt(file).T
    return c11, c22, c33, c12, c13, c23 

@pytest.fixture
def uncertainty_running_average():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/uncertainty_running_average_test.dat")
    uc11, uc22, uc33, uc12, uc13, uc23 = np.loadtxt(file).T
    return uc11, uc22, uc33, uc12, uc13, uc23 

@pytest.fixture
def results_isotropic():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_isotropic_test.dat")
    t, avg, uavg = np.loadtxt(file).T
    return t, avg, uavg

@pytest.fixture
def results_anisotropic():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_anisotropic_test.dat")
    t, avg, uavg = np.loadtxt(file).T
    return t, avg, uavg

class TestGreenKuboIntegral(object):

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_load(self, green_kubo_integral, current):

        integral = green_kubo_integral
        t, Jx, Jy, Jz = current

        assert np.allclose(integral.time, t)
        assert np.allclose(integral.J1, Jx)
        assert np.allclose(integral.J2, Jy)
        assert np.allclose(integral.J3, Jz)

        assert not integral._calculated_caf
        assert not integral._calculated_integral
        assert not integral._calculated_running_average

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_set_time_to_zero(self, green_kubo_integral):

        integral = green_kubo_integral
        integral.time += 10
        integral.set_time_to_zero()
        assert integral.time[0] == 0

    @pytest.mark.parametrize("splits", [1])
    def test_get_correlation_function(self, green_kubo_integral):

        j1 = np.random.random((10, 100))        
        j2 = np.random.random((11, 100))

        with pytest.raises(ValueError):
            _ = green_kubo_integral.get_correlation_function(j1, j2)

        j1 = np.random.random((5, 50))
        j2 = np.random.random((5, 50))

        results = green_kubo_integral.get_correlation_function(j1, j2)

        assert results.shape == (49, 2)

        j1 = np.random.random(50)
        j2 = np.random.random(50)

        results = green_kubo_integral.get_correlation_function(j1, j2)

        assert results.shape == (49, 2)

    @pytest.mark.parametrize("splits", [1])
    def test_analysis_sequence(self, green_kubo_integral):

        with pytest.raises(RuntimeError):
            green_kubo_integral._calculated_caf = False
            green_kubo_integral._calculate_integral()
        with pytest.raises(RuntimeError):
            green_kubo_integral._calculated_caf = False
            green_kubo_integral._calculated_integral = False
            green_kubo_integral._calculate_running_average()
        with pytest.raises(RuntimeError):
            green_kubo_integral._calculated_caf = True
            green_kubo_integral._calculated_integral = False
            green_kubo_integral._calculate_running_average()
        with pytest.raises(RuntimeError):
            green_kubo_integral._calculated_caf = False
            green_kubo_integral._calculated_integral = True
            green_kubo_integral._calculate_running_average()



    @pytest.mark.parametrize("splits", [1])
    def test_calculate_correlation(self, green_kubo_integral, results_correlation, uncertainty_correlation):

        integral = green_kubo_integral
        integral._calculate_correlation()

        assert integral._calculated_caf

        c11, c22, c33, c12, c13, c23 = results_correlation
        uc11, uc22, uc33, uc12, uc13, uc23 = uncertainty_correlation

        # Assert the values

        assert np.allclose(c11, integral.caf["11"][:, 0])
        assert np.allclose(c22, integral.caf["22"][:, 0])
        assert np.allclose(c33, integral.caf["33"][:, 0])
        assert np.allclose(c12, integral.caf["12"][:, 0])
        assert np.allclose(c13, integral.caf["13"][:, 0])
        assert np.allclose(c23, integral.caf["23"][:, 0])

        # Assert the uncertainties

        assert np.allclose(uc11, integral.caf["11"][:, 1])
        assert np.allclose(uc22, integral.caf["22"][:, 1])
        assert np.allclose(uc33, integral.caf["33"][:, 1])
        assert np.allclose(uc12, integral.caf["12"][:, 1])
        assert np.allclose(uc13, integral.caf["13"][:, 1])
        assert np.allclose(uc23, integral.caf["23"][:, 1])

    @pytest.mark.parametrize("splits", [1])
    def test_calculate_integral(self, green_kubo_integral, results_integral, uncertainty_integral):

        integral = green_kubo_integral
        integral._calculate_correlation()
        assert integral._calculated_caf
        integral._calculate_integral()
        assert integral._calculated_integral

        c11, c22, c33, c12, c13, c23 = results_integral
        uc11, uc22, uc33, uc12, uc13, uc23 = uncertainty_integral

        # Assert the values

        assert np.allclose(c11, integral.cumulative_integral["11"][:, 0])
        assert np.allclose(c22, integral.cumulative_integral["22"][:, 0])
        assert np.allclose(c33, integral.cumulative_integral["33"][:, 0])
        assert np.allclose(c12, integral.cumulative_integral["12"][:, 0])
        assert np.allclose(c13, integral.cumulative_integral["13"][:, 0])
        assert np.allclose(c23, integral.cumulative_integral["23"][:, 0])

        # Assert the uncertainties

        assert np.allclose(uc11, integral.cumulative_integral["11"][:, 1])
        assert np.allclose(uc22, integral.cumulative_integral["22"][:, 1])
        assert np.allclose(uc33, integral.cumulative_integral["33"][:, 1])
        assert np.allclose(uc12, integral.cumulative_integral["12"][:, 1])
        assert np.allclose(uc13, integral.cumulative_integral["13"][:, 1])
        assert np.allclose(uc23, integral.cumulative_integral["23"][:, 1])

    @pytest.mark.parametrize("splits", [1])
    def test_calculate_running_average(self, green_kubo_integral, results_running_average, uncertainty_running_average):

        integral = green_kubo_integral
        integral._calculate_correlation()
        assert integral._calculated_caf
        integral._calculate_integral()
        assert integral._calculated_integral
        integral._calculate_running_average()
        assert integral._calculated_running_average

        c11, c22, c33, c12, c13, c23 = results_running_average
        uc11, uc22, uc33, uc12, uc13, uc23 = uncertainty_running_average

        # Assert the values

        assert np.allclose(c11, integral.running_average["11"][:, 0])
        assert np.allclose(c22, integral.running_average["22"][:, 0])
        assert np.allclose(c33, integral.running_average["33"][:, 0])
        assert np.allclose(c12, integral.running_average["12"][:, 0])
        assert np.allclose(c13, integral.running_average["13"][:, 0])
        assert np.allclose(c23, integral.running_average["23"][:, 0])

        # Assert the uncertainties

        assert np.allclose(uc11, integral.running_average["11"][:, 1])
        assert np.allclose(uc22, integral.running_average["22"][:, 1])
        assert np.allclose(uc33, integral.running_average["33"][:, 1])
        assert np.allclose(uc12, integral.running_average["12"][:, 1])
        assert np.allclose(uc13, integral.running_average["13"][:, 1])
        assert np.allclose(uc23, integral.running_average["23"][:, 1])

    
    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_analyze(self, green_kubo_integral):

        green_kubo_integral.analyze()
        assert green_kubo_integral._calculated_caf
        assert green_kubo_integral._calculated_integral
        assert green_kubo_integral._calculated_running_average


    @pytest.mark.parametrize("splits", [1])
    def test_isotropic_conductivity(self, green_kubo_integral, results_isotropic):

        green_kubo_integral.analyze()

        for v1, v2 in zip(results_isotropic, green_kubo_integral.get_isotropic_running_average()):

            assert np.allclose(v1, v2)

    @pytest.mark.parametrize("splits", [1])
    def test_anisotropic_conductivity(self, green_kubo_integral, results_anisotropic):

        green_kubo_integral.analyze()
        
        for v1, v2 in zip(results_anisotropic, green_kubo_integral.get_anisotropic_running_average()):

            assert np.allclose(v1, v2)

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_get_average_correlation_function(self, green_kubo_integral):


        def capped_powerset(iterable):
            """Returns the set of all sets in the iterable that contain at least two elements
            """
            s = list(iterable)
            return list(chain.from_iterable(combinations(s, r) for r in range(2, len(s)+1)))

        keys = ["11", "22", "33", "12", "13", "23"]

        for key_set in capped_powerset(keys):

            t, avg, unc = green_kubo_integral.get_average_correlation_function(key_set)

            assert t.shape == avg.shape
            assert avg.shape == unc.shape

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_specific_correlations(self, green_kubo_integral):

        _, _ , _ = green_kubo_integral.get_isotropic_correlation_function()
        _, _ , _ = green_kubo_integral.get_anisotropic_correlation_function()
       


        
    


# Copyright (c) 2024 The KUTE contributors

try:
    import openmm
except ImportError:
    raise ImportError("In order to use the submodule reporters OpenMM is needed."
    " You can install it following this guide: http://docs.openmm.org/latest/userguide/application/01_getting_started.html#installing-openmm")

from ._comvelocityreporter import COMVelocityReporter
from ._electriccurrentreporter import ElectricCurrentReporter
from ._pressuretensorreporter import PressureTensorReporter
from ._spatialelectriccurrentreporter import SpatialElectricCurrentReporter
__all__ = ["COMVelocityReporter", "ElectricCurrentReporter", "PressureTensorReporter", "SpatialElectricCurrentReporter"]


import sys
import kute.routines._calculate_transport_coefficient as calculate_transport_coefficient
import kute.routines._calculate_microscopic_current as calculate_microscopic_current
import kute.routines._join_currents as join_currents

routines = {"calculate_transport_coefficient": calculate_transport_coefficient, 
            "calculate_microscopic_current": calculate_microscopic_current,
            "join_currents": join_currents}

def main():
    STRING_AVAILABLE = f"Available routines: {list(routines.keys())}"
    if len(sys.argv) == 1:
        print(STRING_AVAILABLE)
        return
    if sys.argv[1] in routines:
        sys.argv[0] = "kute " + sys.argv[1]
        name = sys.argv.pop(1)
        routines[name].main()
    else:
        print(STRING_AVAILABLE)
        sys.exit(1)
        
if __name__ == "__main__":
    main()


# Copyright (c) 2024 The KUTE contributors

import os 
import pytest
import numpy as np

import h5py
import MDAnalysis as mda
from kute import __version__
import getpass
from kute.analysis import ElectricCurrent, COMVelocity, PressureTensor
from MDAnalysis.analysis.base import Results

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]


@pytest.fixture
def universe():
    topo = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")
    traj = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")
    return mda.Universe(topo, traj)

@pytest.fixture
def traj_length():
    topo = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")
    traj = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")

    return len(mda.Universe(topo, traj).trajectory)

@pytest.fixture
def n_atoms():
    topo = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")
    traj = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")

    return len(mda.Universe(topo, traj).atoms)

@pytest.fixture
def n_residues():
    topo = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")
    traj = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")

    return len(mda.Universe(topo, traj).residues)

@pytest.fixture
def resnames():
    topo = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")
    traj = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")

    return np.unique(mda.Universe(topo, traj).residues.resnames)

@pytest.fixture
def results_current():

    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_current_test.h5")
    return h5py.File(file)

@pytest.fixture
def results_com_velocity():

    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_com_velocity_test.h5")
    return h5py.File(file, "r")

@pytest.fixture
def results_pressure_tensor():

    file = os.path.join(ACTUAL_PATH, "files_for_tests/results/results_pressure_tensor_test.h5")
    return h5py.File(file, "r")

class TestElectricCurrent(object):

    def test_init(self, universe):
        
        current = ElectricCurrent(universe)

        assert current.u == universe
        assert current._trajectory == universe.trajectory
        assert current.filename == "current.h5"
        assert current.results == Results()

        current = ElectricCurrent(universe, "test.h5")

        assert current.u == universe
        assert current._trajectory == universe.trajectory
        assert current.filename == "test.h5"
        assert current.results == Results()


    def test_analyze(self, universe, results_current, traj_length, n_atoms):

        ## Initialize analysis 

        current = ElectricCurrent(universe, filename=os.path.join(ACTUAL_PATH, "temp_file.h5"))
        current.run(verbose=False)

        ## Check if internal variables make sense

        assert len(current._weights) == n_atoms
        assert current.results.current.shape == (traj_length, 3)

        ## Check if the file was correctly saved

        assert os.path.isfile(os.path.join(ACTUAL_PATH, "temp_file.h5"))

        ## Check for metadata information

        with h5py.File(os.path.join(ACTUAL_PATH, "temp_file.h5"), "r") as f:

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["electric_current"] == "e * A / ps"
            
            assert np.allclose(np.array(f["timeseries/time"]), np.array(results_current["timeseries/time"]))
            assert np.allclose(np.array(f["timeseries/current"]), np.array(results_current["timeseries/current"]))
            calculated_current_step_1 =  np.array(f["timeseries/current"])[0]

       ## Calculate for first frame manually

        masses = universe.atoms.masses
        residue_masses = np.array([ a.residue.mass for a in universe.atoms ])
        residue_charges = np.array([ a.residue.charge for a in universe.atoms ])
        weights = masses * residue_charges / residue_masses
        velocities = universe.atoms.velocities

        true_current_step_1 = np.sum(weights[:, np.newaxis] * velocities, axis=0)

        assert np.allclose(true_current_step_1, calculated_current_step_1)

        os.remove(os.path.join(ACTUAL_PATH, "temp_file.h5"))

class TestPressureTensor(object):

    def test_init(self, universe):
        
        pressure = PressureTensor(universe)

        assert pressure.u == universe
        assert pressure._trajectory == universe.trajectory
        assert pressure.filename == "pressure_tensor.h5"
        assert pressure.results == Results()

        pressure = PressureTensor(universe, "test.h5")

        assert pressure.u == universe
        assert pressure._trajectory == universe.trajectory
        assert pressure.filename == "test.h5"
        assert pressure.results == Results()

    def test_analyze(self, universe, results_pressure_tensor, traj_length):

        ## Initialize analysis 

        pressure = PressureTensor(universe, filename=os.path.join(ACTUAL_PATH, "temp_file.h5"))
        pressure.run(verbose=False)

        ## Check if internal variables make sense

        assert pressure.results.off_diagonal.shape == (traj_length, 3)

        ## Check if the file was correctly saved

        assert os.path.isfile(os.path.join(ACTUAL_PATH, "temp_file.h5"))

        ## Check for metadata information

        with h5py.File(os.path.join(ACTUAL_PATH, "temp_file.h5"), "r") as f:

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["pressure_tensor"] == "Pa"
            
            assert np.allclose(np.array(f["timeseries/time"]), np.array(results_pressure_tensor["timeseries/time"]))
            assert np.allclose(np.array(f["timeseries/pressure_tensor"]), np.array(results_pressure_tensor["timeseries/pressure_tensor"]), rtol=1e-2)
            calculated_pressure_step_1 =  np.array(f["timeseries/pressure_tensor"])[0]

       ## Calculate for first frame manually

        masses = universe.atoms.masses * 1.66054e-27
        positions = universe.atoms.positions * 1e-10
        velocities = universe.atoms.velocities * 1e-10 / 1e-12
        forces = universe.atoms.forces * 1e3 / (1e-10 * 6.02214076e23)
        volume = universe.dimensions[:3].prod() * 1e-30

        pxy = np.sum(masses * velocities[:, 0] * velocities[:, 1]) / volume + np.sum(positions[:, 0] * forces[:, 1]) / volume
        pxz = np.sum(masses * velocities[:, 0] * velocities[:, 2]) / volume + np.sum(positions[:, 0] * forces[:, 2]) / volume
        pyz = np.sum(masses * velocities[:, 1] * velocities[:, 2]) / volume + np.sum(positions[:, 1] * forces[:, 2]) / volume

        off_diagonal = np.array([pxy, pxz, pyz])

        assert np.allclose(off_diagonal, calculated_pressure_step_1)


class TestCOMVelocity(object):

    def test_init(self, universe):
        
        vel = COMVelocity(universe)

        assert vel.u == universe
        assert vel._trajectory == universe.trajectory
        assert vel.filename == "com_velocity.h5"
        assert vel.results == Results()

        vel = COMVelocity(universe, "test.h5")

        assert vel.u == universe
        assert vel._trajectory == universe.trajectory
        assert vel.filename == "test.h5"
        assert vel.results == Results()

    def test_analyze(self, universe, results_com_velocity, traj_length, n_atoms, n_residues, resnames):

        vel = COMVelocity(universe, filename=os.path.join(ACTUAL_PATH, "temp_file.h5"))
        vel.run(verbose=False)

        ## Check if internal variables make sense

        assert vel._matrix.shape == (n_residues, n_atoms)
        assert vel.results.com_vel.shape == (traj_length, n_residues, 3)

        ## Check for metadata information

        with h5py.File(os.path.join(ACTUAL_PATH, "temp_file.h5"), "r") as f:

            # General information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["com_velocities"] == "A / ps"

            # Residue identificators

            for res in resnames:

                assert res in f["residues"].keys()
                assert np.all(np.array(f[f"residues/{res}"]) == np.array(results_com_velocity[f"residues/{res}"]))
                assert len(np.array(f[f"residues/{res}"])) == len(universe.select_atoms(f"resname {res}").residues)

            # Check result

            assert np.allclose(np.array(f["timeseries/time"]), np.array(results_com_velocity["timeseries/time"]))
            assert np.allclose(np.array(f["timeseries/com_velocities"]), np.array(results_com_velocity["timeseries/com_velocities"]))

        ## Manually calculate the result for the first frame, and the for the first residue of each kind

        for resname in resnames:

            residue = universe.residues[np.where(universe.residues.resnames == resname)[0][0]]
            true_velocity = np.zeros(3)

            for atom in residue.atoms:

                true_velocity += atom.mass / residue.mass * atom.velocity
            
            with h5py.File(os.path.join(ACTUAL_PATH, "temp_file.h5"), "r") as f:
                
                index = np.array(f[f"residues/{resname}"])[0]
                calculated_velocity = np.array(f[f"timeseries/com_velocities"])[0, index, :]

            assert np.allclose(true_velocity, calculated_velocity)
            

        os.remove(os.path.join(ACTUAL_PATH, "temp_file.h5"))
